/*
 Das Programm "BoundingBox" erzeugt um alle selektierten Elemente (Line, Kreise, Bgen und Polylinien) 
 ein umschliesendes Rechteck (BoundingBox). Als Aufrufparameter kann zustzlich ein Offset
 bergeben, um diesen Offsetwert wird das umgebened Rechteck allseitig vergrert. 

 Parameter[1]: Flieskommazahl (Offset)
 
 
 Version 0.1																	
 JonCarpenter	
 (C) 2012-08-12 
 Lizenz: GPL	 

*/
#pragma StackSize 300000

#include "std.h"
#include "megatyp.h"
#include "megacad.h"

#ifndef NOMINMAX
  #ifndef MAX
    #define MAX(a,b) (((a) > (b)) ? (a) : (b))
  #endif
  #ifndef MIN
    #define MIN(a,b) (((a) < (b)) ? (a) : (b))
  #endif
  #define NOMINMAX
#endif  /* NOMINMAX */

#define PI DEF_PI

#define MAXPIDS 65534	// 65535 Elemente sind maximal selektierbar

/**************************************************************************************************/

typedef struct
{
    t_entity  	entity;				// Elementstrucktur muss erstes Element sein
    ushort 		elmnum;  			// Zhler fr die Anzahl der Elemet-ID's
	ulong* 		pElmIDs; 			// Zeiger auf Array mit Element-ID's
	t_rect* 	pBoundingBoxAll;	// Variable fr umgebendes Rechteck aller Elemente 
} t_data;

typedef struct
{
    ushort 		usElmNum;  			// Anzahl der Elemet ID's im Array
	ulong*		pElmIDs; 			// Zeiger auf Array von Element ID's
} t_dodraw;


/**************************************************************************************************/

// erforderlich da Bedingungsoperator in C-Interpreter keine Variablen mit Punktoperator mchte
// die Funktion "Min" ermittelt aus den zwei bergebenen double Werten den kleinsten und 
// gibt diesen zurck
// a			- Wert 1 vom Typ double
// b			- Wert 2 vom Typ double
// Rckgabewert	- Minimum vom Typ double
double Min(double a, double b)
{
	return(MIN(a, b));
}

// erforderlich da Bedingungsoperator in C-Interpreter keine Variablen mit Punktoperator mchte
// die Funktion "Max" ermittelt aus den zwei bergebenen double Werten den grten und 
// gibt diesen zurck
// a			- Wert 1 vom Typ double
// b			- Wert 2 vom Typ double
// Rckgabewert	- Maximum vom Typ double
double Max(double a, double b)
{
	return(MAX(a, b));
}

// die Funktion "BoundingBoxLine" ermittelt das umschliesende Rechteck der bergebenen Linie 
// boundingBox	-	Verweis auf umschliesendes Rechteck vom Typ t_rect 
// line			-	Verweis auf eine Linie vom Typ t_line
// Rckgabewert -	keiner
void BoundingBoxLine(
		t_rect* boundingbox,
		t_line* line)
{
	boundingbox->xmin = Min(line->x1, line->x2);
	boundingbox->ymin = Min(line->y1, line->y2);				
	boundingbox->xmax = Max(line->x1, line->x2);
	boundingbox->ymax = Max(line->y1, line->y2);	

	return;
}

// die Funktion "BoundingBoxCircle" ermittelt das umschliesende Rechteck des bergebenen Kreises
// boundingBox	-	Verweis auf umschliesendes Rechteck vom Typ t_rect 
// circle		-	Verweis aus einen Kreis vom Typ t_circle
// Rckgabewert -	keiner
void BoundingBoxCircle(
		t_rect* 	boundingBox,
		t_circle* 	circle)
{
	boundingBox->xmin = circle->xm - circle->rd;
	boundingBox->ymin = circle->ym - circle->rd;
	boundingBox->xmax = circle->xm + circle->rd;
	boundingBox->ymax = circle->ym + circle->rd;

	return;
}

// Hilfsfunktion fr KreisbogenBoundingBox
// die Funktion "IncludeAngle" prft ob die X-, Y-Koordninaten des Winkels "angle" schon als
// Minima und Maxima der "boundingBox" enthalten sind, wenn nicht werden diese hinzugefgt. 
// boundingBox	-	Verweis auf umschliesendes Rechteck vom Typ t_rect 
// angle		-	Prfwinkel vom Typ double
// Rckgabewert -	keiner
void IncludeAngle(
		t_rect*	boundingBox, 
		double	angle)
{
	double x = cos(angle);
	double y = sin(angle);
	boundingBox->xmin = Min(boundingBox->xmin, x);
	boundingBox->ymin = Min(boundingBox->ymin, y);
	boundingBox->xmax = Max(boundingBox->xmax, x);
	boundingBox->ymax = Max(boundingBox->ymax, y);
	
	return;
}

// Hilfsfunktion fr Kreisbogenwinkel
// die Funktion "CheckAngle" prft ob der Prfwinkel zwischen Start- und Endwinkel liegt
// und wenn ja, dann wird dieser Wert in die BoundingBox einbezogen
// boundingBox	-	Verweis auf umschliesendes Rechteck vom Typ t_rect 
// as 			-	Startwinkel vom Typ double
// ae			-	Endwinkel vom Typ double
// angle		-	Prfwinkel vom Typ double
// Rckgabewert -	keiner
void CheckAngle(
		t_rect* boundingBox, 
		double as, 
		double ae, 
		double angle)
{
	if (as <= angle && ae >= angle) IncludeAngle(boundingBox, angle);

   return;
}

  

// die Funktion "BoundingBoxCirArc" ermittelt das umschliesende Rechteck des bergebenen Kreisbogens
// boundingBox	-	Verweis auf umschliesendes Rechteck vom Typ t_rect 
// cirarc		-	Verweis auf Kreisbogenwerte von Typ t_c_arc
// Rckgabewert -	keiner
void BoundingBoxCirArc(
		t_rect*		boundingBox,
		t_c_arc* 	cirarc)
{
	// Start-, End- und Differenzwinkel
	double as = cirarc->a1;		//Startwinkel
	double ae = cirarc->a2;		//Endwinkel

	// Umrechnung der eigenwilligen MegaCAD Winkelwerte
	// I und II Quadrat positiv und III und IV negativ (von X-Achse aus betrachtet)	
	if(as < 0) as = 2*PI + as;
	if(ae < 0) ae = 2*PI + ae;
	// Wenn EndWinkel kleiner als StartWinkel, dann Endwinkel + 360 (2*PI)
	if(ae < as) ae = ae + 2.0 * PI;
	
	IncludeAngle(boundingBox, as);
	IncludeAngle(boundingBox, ae);
	CheckAngle(boundingBox, as, ae, +0.0 * PI);
	CheckAngle(boundingBox, as, ae, +0.5 * PI);
	CheckAngle(boundingBox, as, ae, +1.0 * PI);
	CheckAngle(boundingBox, as, ae, +1.5 * PI);
	CheckAngle(boundingBox, as, ae, +2.0 * PI);
	CheckAngle(boundingBox, as, ae, +2.5 * PI);
	CheckAngle(boundingBox, as, ae, +3.0 * PI);
	CheckAngle(boundingBox, as, ae, +3.5 * PI);
	CheckAngle(boundingBox, as, ae, +4.0 * PI);
	
	// BoundigBox des Einheitskreises um Radius skalieren und um Mittelpunktskoordinaten verschieben
	boundingBox->xmin = cirarc->xm + boundingBox->xmin * cirarc->rd;
	boundingBox->ymin = cirarc->ym + boundingBox->ymin * cirarc->rd;
	boundingBox->xmax = cirarc->xm + boundingBox->xmax * cirarc->rd;
	boundingBox->ymax = cirarc->ym + boundingBox->ymax * cirarc->rd;
	
	return;
}

// die Funktion "BoundingBoxPolyln" ermittelt das umschliesende Rechteck der bergebenen Polyline
// boundingBox	-	Verweis auf umschliesendes Rechteck vom Typ t_rect 
// cirarc		-	Verweis auf Polyline von Typ t_polyln
// Rckgabewert -	keiner
void BoundingBoxPolyln(
		t_rect*		boundingBox,
		t_polyln*	polyln)
{
	boundingBox->xmin = polyln->rec[0];
	boundingBox->ymin = polyln->rec[1];
	boundingBox->xmax = polyln->rec[2];
	boundingBox->ymax = polyln->rec[3];	
	
	return;
}

// die Funktion "BoundingBoxEntity" ermittelt das umschliesende Rechteck des jeweiligen Elementtyps
// boundingBox	-	Verweis auf umschliesendes Rechteck vom Typ t_rect 
// pEntity		-	Verweis auf Element von Typ t_entity
// Rckgabewert -	keiner
void BoundingBoxEntity(
		t_rect*		pBoundingBox,
		t_entity*	pEntity)
{	
	switch((ushort) pEntity->type)
	{
	// Linie auswerten
	case E_LINE:	BoundingBoxLine(pBoundingBox, (t_line*) pEntity->data);
					break;
	// Kreis auswerten
	case E_CIRCLE:	BoundingBoxCircle(pBoundingBox, (t_circle*) pEntity->data);
					break;
	// Bogen auswerten			
	case E_C_ARC:	BoundingBoxCirArc(pBoundingBox, (t_c_arc*) pEntity->data);
					break;
	// Polyline auswerten			
	case E_POLYLN:	BoundingBoxPolyln(pBoundingBox, (t_polyln*) pEntity->data);
					break;
	}
	
	return;
}


// die Funktion "DoDraw" zeichnet die bergebenen Element auf dem Bildschirm
// dowraw 		- zu zeichnende Elemente vom Typ t_dodraw
// Rckgabewert	- keiner;
void DoDraw(
		t_dodraw* dodraw)
{
 	DrawEntities(dodraw->pElmIDs, dodraw->usElmNum);
	return;
}

// die Funktion "CreateRect" erstellt ein umschieendes Rechteck erstellen
// data			- Verweis auf Daten vom Typ t_data
// offset		- Offset vom Typ double fr umschliesendes Rechteck 
// Rckgabewert	- keiner;
void CreateRect(
		t_data* data,
		double offset)
{
	t_line   	line[8];
	t_dodraw 	dodraw;
	short		ret = 0;
	
	SetFuncText("umschieendes Rechteck erstellen - Rechteck (blau), Elemente blinken");
	MouseHelp("erstellen","verwerfen",HLP_INV(0,0,0));
	
	// Rechteck mit offset
	// waagerecht unten
	line[0].x1 = data->pBoundingBoxAll->xmin - offset;
	line[0].x2 = data->pBoundingBoxAll->xmax + offset;
	line[0].y1 = data->pBoundingBoxAll->ymin - offset;
	line[0].y2 = data->pBoundingBoxAll->ymin - offset;
	// senkrecht rechts
	line[1].x1 = data->pBoundingBoxAll->xmax + offset;
	line[1].x2 = data->pBoundingBoxAll->xmax + offset;
	line[1].y1 = data->pBoundingBoxAll->ymin - offset;
	line[1].y2 = data->pBoundingBoxAll->ymax + offset;	
	// waagerecht oben
	line[2].x1 = data->pBoundingBoxAll->xmax + offset;
	line[2].x2 = data->pBoundingBoxAll->xmin - offset;
	line[2].y1 = data->pBoundingBoxAll->ymax + offset;
	line[2].y2 = data->pBoundingBoxAll->ymax + offset;	
	// senkrecht links
	line[3].x1 = data->pBoundingBoxAll->xmin - offset;
	line[3].x2 = data->pBoundingBoxAll->xmin - offset;
	line[3].y1 = data->pBoundingBoxAll->ymax + offset;
	line[3].y2 = data->pBoundingBoxAll->ymin - offset;	

	// Rechteck ohne offset
	// waagerecht unten
	line[4].x1 = data->pBoundingBoxAll->xmin;
	line[4].x2 = data->pBoundingBoxAll->xmax;
	line[4].y1 = data->pBoundingBoxAll->ymin;
	line[4].y2 = data->pBoundingBoxAll->ymin;
	// senkrecht rechts
	line[5].x1 = data->pBoundingBoxAll->xmax;
	line[5].x2 = data->pBoundingBoxAll->xmax;
	line[5].y1 = data->pBoundingBoxAll->ymin;
	line[5].y2 = data->pBoundingBoxAll->ymax;	
	// waagerecht oben
	line[6].x1 = data->pBoundingBoxAll->xmax;
	line[6].x2 = data->pBoundingBoxAll->xmin;
	line[6].y1 = data->pBoundingBoxAll->ymax;
	line[6].y2 = data->pBoundingBoxAll->ymax;	
	// senkrecht links
	line[7].x1 = data->pBoundingBoxAll->xmin;
	line[7].x2 = data->pBoundingBoxAll->xmin;
	line[7].y1 = data->pBoundingBoxAll->ymax;
	line[7].y2 = data->pBoundingBoxAll->ymin;		

	// Rechteck mit Offset temporr zeichnen (hellblau)
	DrawInv(E_LINE, line[0], LMODE(1,9,6), 0);
	DrawInv(E_LINE, line[1], LMODE(1,9,6), 0);
	DrawInv(E_LINE, line[2], LMODE(1,9,6), 0);
	DrawInv(E_LINE, line[3], LMODE(1,9,6), 0);
	
	if(offset != 0)
	{
		// Rechteck ohne Offset temporr zeichnen (hellgruen)
		DrawInv(E_LINE, line[4], LMODE(2,10,6), 0);
		DrawInv(E_LINE, line[5], LMODE(2,10,6), 0);
		DrawInv(E_LINE, line[6], LMODE(2,10,6), 0);
		DrawInv(E_LINE, line[7], LMODE(2,10,6), 0);
	}
	
	// selektierte Elemente blinken lassen
	dodraw.usElmNum = data->elmnum;
	dodraw.pElmIDs  = data->pElmIDs;
	ret = Blink(DoDraw, &dodraw);	

	EraseInv();

	if( ret != 0)
	{
		// Linien erzeugen Attribute aus Leiste, Koordinaten UCS, am Ende
		// (Typ, Attribute, Daten, Besitzer, Koordinaten, anhngen_an)
		CreateEntity(E_LINE, NULL, line[0], NULL, NULL, 0);
		CreateEntity(E_LINE, NULL, line[1], NULL, NULL, 0);
		CreateEntity(E_LINE, NULL, line[2], NULL, NULL, 0);
		CreateEntity(E_LINE, NULL, line[3], NULL, NULL, 0);
		savecount();
	}
	
	return(0);
}

// die Funktion "SelectElement" wird fr jedes zugelassene selektierte Element aufgerufen
// von dieses wird die jeweis die BoundingBox ermittelt und ggf. die Ninimalen und Maximalen Werte
// fr die BoundingBox der gesammten Elementen verwendet.
// pArg 		- Verweis auf Stuktur vom Typ t_data
// Rckgabewert - Schleife fortsetzen oder abbrechen
short SelectElement(
		void * pArg )
{
	t_data* pData;
	t_rect boundingBoxElement;
	
	pData = (t_data*) pArg;
	
	pData->pElmIDs[pData->elmnum] = pData->entity.id;
	pData->elmnum++;

	boundingBoxElement.xmin = 0;
	boundingBoxElement.ymin = 0;
	boundingBoxElement.xmax = 0;
	boundingBoxElement.ymax = 0;
	
	// BoundingBox des selektierten Elements ermitteln
	BoundingBoxEntity(&boundingBoxElement, pData->entity);
	
	// vergleich der MinMax-Werte von der Element BoundingBox mit den Werten der Box aller Elemente
	// ggf. Wert Extremwerte eintragen
	pData->pBoundingBoxAll->xmin = Min(pData->pBoundingBoxAll->xmin, boundingBoxElement.xmin);
	pData->pBoundingBoxAll->xmax = Max(pData->pBoundingBoxAll->xmax, boundingBoxElement.xmax);
	pData->pBoundingBoxAll->ymin = Min(pData->pBoundingBoxAll->ymin, boundingBoxElement.ymin);
	pData->pBoundingBoxAll->ymax = Max(pData->pBoundingBoxAll->ymax, boundingBoxElement.ymax);
	
	if(pData->elmnum >= MAXPIDS)
	{
		Message("Fehler", "Zu viele Elemente selektiert!", "", "", "OK", 3);
		return(BREAK_LOOP);
	}
	// selektierte Elemente invers zeichnen (rot)
	EraseInv();
	DrawInvIDvec(pData->pElmIDs, pData->elmnum, LMODE(1,12,6));

	return(CONTINUE_LOOP);
}



// die Funktion "SelectElements" erstellt eine selektionsschleife fr ausgewhlte Elementtypen
// pData 		-  Verweis auf Stuktur vom Typ t_data
// Rckgabewert - Schleife fortsetzen oder abbrechen
void SelectElements(
		t_data* pData)
{
    t_mat44  ecs;
	ulong	filter1 = 0x0;	//0xffffffff;
	ulong	filter2 = 0x0;	//0xffffffff;

    // Selektionskriterium Linien Kreise Bgen Polylinien
	filter1 = ((1<<E_LINE) | (1<<E_CIRCLE) | (1<<E_C_ARC) | (1<<E_POLYLN)); 

	SetFuncText("umschieendes Rechteck erstellen - Elemente auswhlen");
	MouseHelp("selektieren","abbrechen",HLP_INV(0,0,0));
	
    // Selektionsschleife bei jedem gefundenen Element wird SelID aufgerufen
    SelectInit(filter1, filter2);
    SelectLoop(SelectElement, pData, GetVarPtr(), &ecs, VD_CNT, FALSE);
	
	// Temporre Inverse Elemente lschen
	EraseInv();
	
	return;
}


/**********************************************************************/
// Haupprogramm wird von MegaCAD aufgerufen
short MegaMain(
        char *filename,
        char *args )
{
	
	ulong 	 	elmIDs[MAXPIDS]; 		// Array fr Element ID's
	t_data	 	data;					// Struktur fr Daten
	t_dodraw 	dodraw; 				// Struktur fr Anzeigelemente (Blinken)
	t_rect		boundingBoxAll;			// Struktur fr umschliesendes Rechteck
	double		offset = 0;

	data.pBoundingBoxAll = &boundingBoxAll;
	data.pElmIDs 		 = elmIDs;
	data.elmnum  		 = 0;
	
	// jeweils mit mininmalen oder maximalen Gegeteil initialisieren
	boundingBoxAll.xmin = +1.797693e+307;
	boundingBoxAll.ymin = +1.797693e+307;
	boundingBoxAll.xmax = -1.797693e+307; 
	boundingBoxAll.ymax = -1.797693e+307; 

	offset = atof(args);
	
	SetCharSet(GT_ANSI);
	
	// Funktionstext und Maushilfe in der statuszeile anzeigen
	SetFuncText("Rechteck um Elemente zeichnen");
	
	// prfen ob 2D Moduns aktiv ist, da kein Koorninatensytem ausgewertet wird!
	if(IsMod2D())
	{
		// Elemente selektieren
		SelectElements(&data);

		
		// wurden Elemente selektiert, wenn ja Rechteck erstellen
		if(data.elmnum != 0)
		{
			// umschlieendes Rechteck erstellen (Boundingbox)
			CreateRect(&data, offset);
		}
	}
	else
	{
		Message("Fehler - 2D-Modus nicht aktiv",
				"Rechteck kann nur im 2D-Modus erstellt werden.",
				NULL, NULL, "OK", 3);
	}	
	
	SetFuncText("");	
	
    return(0);
}
